/**
 * This file is part of cs838-dc1
 *
 * cs838-dc1 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 *  cs838-dc1 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with cs838-dc1.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   Copyright 2010 James Hill
 *   */

#include "frame.h"
#include "GeneralException.h"
#include "conceptNamer.h"

#include <QFile>
#include <QTextStream>
#include <QStringList>

#include <iostream>
using namespace std;

Frame::Frame() : data(0), frameSize(0)
{

}

Frame::Frame(const Frame & frame)
{
	data = 0;
	frameSize = 0;
	(*this) = frame;	
}

const Frame & Frame::operator=(const Frame & frame)
{
	if(&frame == this)
		return *this;

	clear();

	// copy the frame size
	frameSize = frame.frameSize;

	// copy the data
	data = new float[frameSize * frameSize];
	for(int i = 0, stop = frameSize * frameSize; i < stop; i++)
		data[i] = frame.data[i];

	// copy the frame name
	frameName = frame.frameName;

	// copy the concept names
	concepts.clear();

	for(unsigned int i = 0, stop = frame.concepts.size(); i < stop; i++)
		concepts.push_back(frame.concepts[i]);

	return *this;	
}

Frame::~Frame()
{
	clear();
}


// load a frame from a csv file
// main diag is zeroed
void Frame::load(const QString & file)
{
	// open the file
	QFile f(file);
	if(!f.open(QIODevice::ReadOnly))
		throwGen("Failed to open file [%s]", file.toAscii().data());
			
	// read each line into a vector
	vector<QString> lines;
	QTextStream ts(&f);
	QString line;
	while(!(line = ts.readLine()).isNull())
	{
		if(line.trimmed() != "")
			lines.push_back(line);
	}
	
	// split each line and make sure it's the correct size	
	for(unsigned int i = 0, stop = lines.size(); i < stop; i++)
	{
		QStringList list = lines[i].split(",");
		if(static_cast<unsigned int>(list.count()) != stop)
			throwGen("Line [%d] has invalid concept count", i + 1);
		
		// also make sure we can convert each value to a float
		bool result = false;
		for(unsigned int j = 0; j < stop; j++)
		{
			list[j].toFloat(&result);
			if(!result)
				throwGen("value at line [%d] index [%d] isn't a valid float", i + 1, j + 1);
		}
	}

	// load up the data
	clear();

	// get the size
	frameSize = lines.size();
	data = new float[frameSize * frameSize];

	// get the data	
	float * d = data;
	for(unsigned int i = 0, stop = lines.size(); i < stop; i++)
	{
		QStringList list = lines[i].split(",");
		for(unsigned int i = 0; i < stop; i++, d++)
			(*d) = list[i].toFloat();
	}

	// load up some concept names
	QString concept = "A";
	for(int i = 0; i < frameSize; i++)
	{
		concepts.push_back(concept);
		nextName(concept);
	}	
	
	// get the file name
	QStringList name = file.split("/");
	frameName = name.back();

	// zero out the main diag
	for(int i = 0; i < frameSize; i++)
		data[i * frameSize + i] = 0.0f;
}

// write a frame to a csv file
// main diag is zeroed
// current name is used as the file name
void Frame::write(const QString & dir)
{
	QString name;
	if(dir == "")
		name = QString("./") + frameName;
	else
		name = dir + "/" + frameName;
	QFile out(name);
	bool open = out.open(QIODevice::WriteOnly);
	if(!open)
		throwGen("Unable to open file [%s] for writing", dir.toAscii().data());

	QTextStream ts(&out);
	for(int l = 0; l < frameSize; l++)
	{
		for(int c = 0; c < frameSize; c++)
		{
			ts << data[l * frameSize + c];
			if(c + 1 == frameSize && l + 1 != frameSize)
				ts << "\n";
			else if(c + 1 != frameSize)
				ts << ",";
		}
	}
	ts.flush();
	out.close();
}

// returns the concept weight between a and b,
// will return 0.0 if a == b
float & Frame::operator()(const QString & a, const QString & b)
{
	int ia, ib;
	unsigned int i, stop;
	for(i = 0, stop = concepts.size(); i < stop; i++)
		if(concepts[i] == a)
			break;
	if(i == stop)
		throwGen("Failed to find concept [%s]", a.toAscii().data());
	ia = i;

	for(i = 0; i < stop; i++)
		if(concepts[i] == b)
			break;
	if(i == stop)
		throwGen("Failed to find concept [%s]", b.toAscii().data());

	ib = i;
	
	return (*this)(ia, ib);
}

const float & Frame::operator()(const QString & a, const QString & b) const
{
	int ia, ib;
	unsigned int i, stop;
	for(i = 0, stop = concepts.size(); i < stop; i++)
		if(concepts[i] == a)
			break;
	if(i == stop)
		throwGen("Failed to find concept [%s]", a.toAscii().data());
	ia = i;

	for(i = 0; i < stop; i++)
		if(concepts[i] == b)
			break;
	if(i == stop)
		throwGen("Failed to find concept [%s]", b.toAscii().data());

	ib = i;
	
	return (*this)(ia, ib);
}

// access as a = row b = column
float & Frame::operator()(int a, int b)
{
	if(a >= frameSize)
		throwGen("index [%d] is out of bounds", a);
	if(b >= frameSize)
		throwGen("indes [%d] is out of bounds", b);

	return data[a * frameSize + b];	
}
	
const float & Frame::operator()(int a, int b) const
{
	if(a >= frameSize)
		throwGen("index [%d] is out of bounds", a);
	if(b >= frameSize)
		throwGen("indes [%d] is out of bounds", b);

	return data[a * frameSize + b];	
}
	

// write the concept names as a vector
void Frame::getConcepts(vector<QString> & names) const
{
	names.clear();
	names = concepts;
}


// return the concept at a given index
const QString & Frame::getConcept(int c) const
{
	if(c >= frameSize)
		throwGen("index [%d] is out of bounds", c);

	return concepts[c];
}

// return the size
int Frame::size() const
{
	return frameSize;
}

// set the current size (no data is carried)
void Frame::setSize(int size)
{
	clear();
	data = new float(size * size);
	frameSize = size;	
	
	// load up some concept names
	QString concept = "A";
	for(int i = 0; i < frameSize; i++)
	{
		concepts.push_back(concept);
		nextName(concept);
	}
}

// return the frame name (file name)
const QString & Frame::name() const
{
	return frameName;
}

// set the current frame name
void Frame::setName(const QString & name)
{
	frameName = name;
}

// return the raw data vector as a const pointer
const float * Frame::raw() const
{
	return data;
}
		
void Frame::clear()
{
	if(data)
	{
		delete [] data;
		data = 0;
	}
	
	frameSize = 0;

	concepts.clear();

	frameName = "";
}

bool framesAreCompatible(const Frame & a, const Frame & b)
{
	if(a.size() != b.size())
		return false;

	vector<QString> aConcepts, bConcepts;
	a.getConcepts(aConcepts);
	b.getConcepts(bConcepts);

	if(aConcepts.size() != bConcepts.size())
		return false;

	for(unsigned int i = 0; i < aConcepts.size(); i++)
		if(aConcepts[i] != bConcepts[i])
			return false;
	
	return true;
}

void checkFrames(const Frame & a, const Frame & b)
{
	if(!framesAreCompatible(a,b))
		throwGen("Frames [%s] and [%s] are not compatible",
			a.name().toAscii().data(),
			b.name().toAscii().data());
}

void subtract(const Frame & a, const Frame & b, Frame & c)
{
	checkFrames(a, b);

	c = a;

	for(int i = 0, stop = a.size(); i < stop; i++)
		for(int j = 0; j < stop; j++)
			c(i,j) = a(i,j) - b(i,j);

}
