/**
 * This file is part of cs838-dc1
 *
 * cs838-dc1 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 *  cs838-dc1 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with cs838-dc1.  If not, see <http://www.gnu.org/licenses/>.
 *
 *   Copyright 2010 James Hill
 *   */

#include "distance_widget.h"

#include <QMouseEvent>
#include <QPaintEvent>
#include <QResizeEvent>
#include <QImage>
#include <QPixmap>
#include <QRect>
#include <QPen>
#include <QBrush>
#include <QPainter>
#include <QPoint>
#include <QSizePolicy>

#include "GeneralException.h"
#include "distance.h"

#include <cmath>
#include <cfloat>

#include <iostream>
using namespace std;

DistanceWidget::DistanceWidget(QWidget * parent) : 
		QWidget(parent), 
		mainFrame(0), 
		mustRedrawImage(true), 
		distName("Simple Distance"),
		nodeDiameter(40)
{
	// make sure we have a valid dist function
	if(getFrameDistanceFunction(distName) == 0)
		throwGen("Failed to find default distance function [%s]", "Simple Distance");

	// build some initial images for drawing		
	pickImage = new QImage(width(), height(), QImage::Format_ARGB32);
	viewImage = new QPixmap(width(), height());

	// set some default colors		
	nodeColor.setRgb(100, 255, 100);	// light green
	nodeOutline.setRgb(0, 255, 0);		// pure green
	mainNodeColor.setRgb(255, 100, 100);	// light red
	mainNodeOutline = nodeOutline;		// pure green

	// we want mouse tracking
	setMouseTracking(true);

	// set some size policy
	setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));
}

DistanceWidget::~DistanceWidget()
{
	// nothing to do, we don't own any memory
}


void DistanceWidget::loadFrames(const vector<Frame*> & frames)
{
	if(frames.size() == 0)
		return;

	frameData = frames;
	mainFrame = frameData[0]; // just start with the first

	// generate picking info
	for(unsigned int i = 1; i <= frameData.size(); i++)
	{
		colorPickFrame[i] = frameData[i - 1];
		framePickColor[frameData[i - 1]] = i;
		hovered.push_back(0);
		selected.push_back(0);	
	}
	
	generateDists();
}
		
void DistanceWidget::setMainFrame(const Frame * frame)
{
	// make sure the frame exists
	unsigned int i = 0;
	for(; i < frameData.size(); i++)
		if(frameData[i] == frame)
			break;

	if(i == frameData.size())
		return;
	
	mainFrame = frameData[i];
	generateDists();

	// update the widget
	mustRedrawImage = true;
	update();
}

void DistanceWidget::setDistName(const QString & name)
{
	// if we can't find the name return
	if(getFrameDistanceFunction(name) == 0)
		return;

	distName = name;
}
	
void DistanceWidget::clear()
{
	mainFrame = 0;
	frameData.clear();
	dists.clear();

	mustRedrawImage = true;
	update();
}
		
void DistanceWidget::shouldUpdate()
{
	mustRedrawImage = true;
	generateDists();
	update();
}
		
void DistanceWidget::getSelectedFrames(vector<Frame*> & selectedFrames)
{
	selectedFrames.clear();
	for(unsigned int i = 0; i < selected.size(); i++)
		if(selected[i] != 0)
			selectedFrames.push_back(frameData[i]);
}
		
void DistanceWidget::loadMinDistance()
{
	// loop through all of the frames setting each to the main frame and
	// calling generateDists, pick the one that has the smallest average
	// distance

	if(frameData.size() == 0)
		return;
	
	float minDistance = FLT_MAX;
	unsigned int minIndex = 0;

	unsigned int i;
	for(i = 0; i < frameData.size(); i++)
	{
		mainFrame = frameData[i];
		generateDists();

		// calculate the average distance
		float average = 0.0f;
		for(unsigned int j = 0; j < dists.size(); j++)
			average += dists[j];
		
		average /= static_cast<float>(dists.size());

		if(average < minDistance)
		{
			minDistance = average;
			minIndex = i;
		}
	}

	mainFrame = frameData[minIndex];
	generateDists();

	mustRedrawImage = true;
	update();	
}

void DistanceWidget::loadMaxDistance()
{
	// loop through all of the frames setting each to the main frame and
	// calling generateDists, pick the one that has the largest average
	// distance

	if(frameData.size() == 0)
		return;
	
	float maxDistance = FLT_MIN;
	unsigned int maxIndex = 0;

	unsigned int i;
	for(i = 0; i < frameData.size(); i++)
	{
		mainFrame = frameData[i];
		generateDists();

		// calculate the average distance
		float average = 0.0f;
		for(unsigned int j = 0; j < dists.size(); j++)
			average += dists[j];
		
		average /= static_cast<float>(dists.size());

		if(average > maxDistance)
		{
			maxDistance = average;
			maxIndex = i;
		}
	}

	mainFrame = frameData[maxIndex];
	generateDists();

	mustRedrawImage = true;
	update();	
}


void DistanceWidget::mouseMoveEvent(QMouseEvent * event)
{
	if(event->x() >= pickImage->width() || event->y() >= pickImage->height() ||
		event->x() < 0 || event->y() < 0)
		return;
	
	unsigned int color = pickImage->pixel(event->x(), event->y()) & 0x00ffffff;

	clearHovered();

	// if it's not zero, look it up in our pick map
	if(color != 0)
	{
		map<QRgb, const Frame *>::iterator iter = colorPickFrame.find(color);
		if(iter == colorPickFrame.end())
		{
			cout << "Failed to find color: " << color << endl;
			return;
		}

		hovered[color - 1] = 1;
	}
	
	update();
}

void DistanceWidget::mouseReleaseEvent(QMouseEvent * event)
{
	if(event->x() >= pickImage->width() || event->y() >= pickImage->height() ||
		event->x() < 0 || event->y() < 0)
		return;

	// first check for left or right mouse click	
	unsigned int color = pickImage->pixel(event->x(), event->y()) & 0x00ffffff;
	
	// if it's not zero, look it up in our pick map
	if(color != 0)
	{
		map<QRgb, const Frame *>::iterator iter = colorPickFrame.find(color);
		if(iter == colorPickFrame.end())
		{
			cout << "Failed to find color: " << color << endl;
			return;
		}

		if(event->button() == Qt::LeftButton)
		{
			selected[color - 1] = !selected[color -1];
		}
	}
	
	update();
}

void DistanceWidget::mouseDoubleClickEvent(QMouseEvent * event)
{
	if(event->x() >= pickImage->width() || event->y() >= pickImage->height() ||
		event->x() < 0 || event->y() < 0)
		return;

	// first check for left or right mouse click	
	unsigned int color = pickImage->pixel(event->x(), event->y()) & 0x00ffffff;
	
	// if it's not zero, look it up in our pick map
	if(color != 0)
	{
		map<QRgb, const Frame *>::iterator iter = colorPickFrame.find(color);
		if(iter == colorPickFrame.end())
		{
			cout << "Failed to find color: " << color << endl;
			return;
		}

		if(event->button() == Qt::LeftButton)
		{
			// set the new main frame
			setMainFrame(frameData[color - 1]);
			clearHovered();
			clearSelected();
		}
	}
	
	update();

}

void DistanceWidget::paintEvent(QPaintEvent *)
{
	static float pi2 = 8.0f * atan(1.0f); // pi

	QPainter painter;

	if(mustRedrawImage)
	{
		// get the image buffers ready
		prepareImageBuffers();
		pickImage->fill(0);
		viewImage->fill(QColor(0, 0, 0));

		// we need to know a couple of things
		// the angle between each node
		float nodeAngle = pi2; 
		if(frameData.size() > 1)
			nodeAngle = pi2 / static_cast<float>(frameData.size() - 1);
		
		// calculate a scaling factor based on the max distance and
		// the width of the widget
		int centerX = width() / 2;
		int centerY = height() / 2;
		int maxPixelValue;
		if(width() > height())
			maxPixelValue = height() - centerY - nodeDiameter / 2;
		else
			maxPixelValue = width() - centerX - nodeDiameter / 2;
		float scaleFactor = static_cast<float>(maxPixelValue) / maxDist;

		float currentAngle = 0.0f;

		// draw the pick image
		painter.begin(pickImage);
		painter.setRenderHint(QPainter::Antialiasing, false);
			// basically walk around the unit circle in steps of nodeAngle
			// untill all frames are drawn
			for(unsigned int i = 0; i < frameData.size(); i++)
			{
				// if this is the main node, just draw it
				if(frameData[i] == mainFrame)
					drawNode(frameData[i], centerX, centerY, &painter, true, true);
				else
				{
					// we have to calculate the unit vector for the current angle.
					float ux = cos(currentAngle);
					float uy = sin(currentAngle);
					ux = ux * dists[i] * scaleFactor;
					uy = uy * dists[i] * scaleFactor;
					drawNode(frameData[i],
							centerX + static_cast<int>(ux),
							centerY + static_cast<int>(uy),
							&painter, 
							false,
							true);	
					
					currentAngle += nodeAngle;
				}
			}
		
		painter.end();

		// draw the view image
		painter.begin(viewImage);
		painter.setRenderHint(QPainter::Antialiasing, true);
			// basically walk around the unit circle in steps of nodeAngle
			// untill all frames are drawn
			int mainNode = -1;
			for(unsigned int i = 0; i < frameData.size(); i++)
			{
				// if this is the main node, just draw it
				if(frameData[i] == mainFrame)
					mainNode = i;
				else
				{
					// we have to calculate the unit vector for the current angle.
					float ux = cos(currentAngle);
					float uy = sin(currentAngle);
					ux = ux * dists[i] * scaleFactor;
					uy = uy * dists[i] * scaleFactor;
						
					// draw a spoke
					QPen pen(QColor(150, 150, 150));
					painter.setPen(pen);
					painter.drawLine(centerX, centerY, 
							centerX + static_cast<int>(ux),
							centerY + static_cast<int>(uy));
					
					drawNode(frameData[i],
							centerX + static_cast<int>(ux),
							centerY + static_cast<int>(uy),
							&painter, 
							false,
							false);	
				
					currentAngle += nodeAngle;
				}
			}

			if(mainNode != -1)
				drawNode(frameData[mainNode], centerX, centerY, &painter, true, false);
		
		painter.end();
		mustRedrawImage = false;
	}

	// draw the pixmap to the screen	
	painter.begin(this);
	painter.drawPixmap(0, 0, *viewImage); 
	painter.setRenderHint(QPainter::Antialiasing, true);

	// draw hovered or selected nodes
	int indexOffset = 0;
	for(unsigned int i = 0, stop = hovered.size(); i < stop; i++)
	{
		if(frameData[i] == mainFrame)
			indexOffset = 1;

		if(hovered[i] == 1 || selected[i] == 1)
		{
			QPen pen(QColor(255, 255, 255));
			pen.setWidth(5);
			painter.setPen(pen);
			QBrush brush;
			painter.setBrush(brush);
			
			int centerX = width() / 2;
			int centerY = height() / 2;
		
			int maxPixelValue;
			if(width() > height())
				maxPixelValue = height() - centerY - nodeDiameter / 2;
			else
				maxPixelValue = width() - centerX - nodeDiameter / 2;
			float scaleFactor = static_cast<float>(maxPixelValue) / maxDist;
		
			float nodeAngle = pi2; 
			if(frameData.size() > 1)
				nodeAngle = pi2 / static_cast<float>(frameData.size() - 1);

			float angle = nodeAngle * static_cast<float>(i - indexOffset);
		
			float ux = cos(angle);
			float uy = sin(angle);
			ux = ux * dists[i] * scaleFactor;
			uy = uy * dists[i] * scaleFactor;
			painter.drawEllipse(centerX + static_cast<int>(ux) - nodeDiameter / 2, 
						centerY + static_cast<int>(uy) - nodeDiameter / 2, 
						nodeDiameter, nodeDiameter);

			// draw a tool tip type thing too
			QRect bound = painter.boundingRect(0, 0, 10000, 10000, Qt::AlignHCenter, frameData[i]->name());
			brush.setStyle(Qt::SolidPattern);
			brush.setColor(QColor(255, 255, 255));
			painter.setBrush(brush);
			QRect labelBox(centerX + static_cast<int>(ux) + nodeDiameter / 2, 
					centerY + static_cast<int>(uy) - nodeDiameter / 2 - bound.height(),
					bound.width(), bound.height());
			painter.drawRect(labelBox);

			pen.setWidth(1);
			pen.setColor(QColor(0, 0, 0));
			painter.setPen(pen);
			painter.drawText(labelBox, Qt::AlignHCenter, frameData[i]->name());
		}
	}
}

void DistanceWidget::resizeEvent(QResizeEvent *)
{
	mustRedrawImage = true;
	update();	
}


void DistanceWidget::generateDists()
{
	dists.clear();

	frameDistance distFunction = getFrameDistanceFunction(distName);

	minDist = FLT_MAX;
	maxDist = FLT_MIN;

	// calculate a diff between each function and the mainFrame
	// the main frame will have zero distance to itself, but we just
	// won't draw it
	for(unsigned int i = 0, stop = frameData.size(); i < stop; i++)
	{
		cout.flush();
		float d = (*distFunction)(*mainFrame, *(frameData[i]));
		dists.push_back(d);
		if(d > maxDist)
			maxDist = d;
		if(d < minDist)
			minDist = d;
	}
}

		
void DistanceWidget::prepareImageBuffers()
{
	// we keep pickImage and viewImage tightly coupled so we only 
	// need looke at the dimensions and address of one
	if(pickImage->width() != width() || pickImage->height() != height())
	{
		delete pickImage;
		delete viewImage;
		pickImage = new QImage(width(), height(), QImage::Format_ARGB32);
		viewImage = new QPixmap(width(), height());
	}
}
		
void DistanceWidget::drawNode(const Frame * frame, int x, int y, QPainter * painter, bool isMainNode, bool isPick)
{
	int dx = x - nodeDiameter / 2;
	int dy = y - nodeDiameter / 2;

	if(isPick)
	{
		// just draw the elipse with the required color
		unsigned int color = framePickColor[frame];
		QColor pickColor((QRgb)color);
		QPen pen(pickColor);
		QBrush brush(pickColor);
		painter->setPen(pen);
		painter->setBrush(brush);
		painter->drawEllipse(dx, dy, nodeDiameter, nodeDiameter);
	}
	else
	{
		// draw the actual node
		QPen pen;
		QBrush brush;

		if(isMainNode)
		{
			brush.setColor(mainNodeColor);
			pen.setColor(mainNodeColor);
		}
		else
		{
			brush.setColor(nodeColor);
			pen.setColor(nodeColor);
		}
		brush.setStyle(Qt::SolidPattern);

		// draw the main body
		painter->setPen(pen);
		painter->setBrush(brush);
		painter->drawEllipse(dx, dy, nodeDiameter, nodeDiameter);
	}
}
		
void DistanceWidget::clearHovered()
{
	for(unsigned int i = 0, stop = hovered.size(); i < stop; i++)
		hovered[i] = 0;
}

void DistanceWidget::clearSelected()
{
	for(unsigned int i = 0, stop = selected.size(); i < stop; i++)
		selected[i] = 0;
}
